@@ -23,18 +23,15 @@ from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
 from sklearn.metrics import f1_score, classification_report, confusion_matrix
 
 
+from typing import Callable
 
 
 
-
-# from counter import Counter
-# from nans import Nans
-
-
-from . import counter
 from . import nans
 from . import encoder
 from . import splitter
+from . import cross_validation
+from . import recover
 
 
 
@@ -42,17 +39,11 @@ class Snaplib__222:
     '''
     data preprocessing library
     '''
-
     ENCODER = encoder.Encoder()
 
 
 
 
-
-
-
-
-
     @staticmethod
     def nan_info(df : pd.DataFrame) -> pd.DataFrame:
         '''
@@ -60,9 +51,6 @@ class Snaplib__222:
         Use case:
         missing_info_df = Snaplib().nan_info(df)
         '''
-
-        if not isinstance(df, pd.core.frame.DataFrame):
-            raise TypeError('The df must be a pandas.core.frame.DataFrame instance.')
         return nans.nan_info(df)
 
 
@@ -75,9 +63,6 @@ class Snaplib__222:
         Use case:
         Snaplib().nan_plot(df)
         '''
-
-        if not isinstance(df, pd.core.frame.DataFrame):
-            raise TypeError('The df must be a pandas.core.frame.DataFrame instance.')
         return nans.nan_plot(df)
 
 
@@ -95,36 +80,21 @@ class Snaplib__222:
         Use case:
         df = Snaplib().cleane(df, target_name, verbose=True)
         '''
-
-        if not isinstance(df, pd.core.frame.DataFrame):
-            raise TypeError('The df must be a pandas.core.frame.DataFrame instance.')
-        if target is not None:
-            if type(target) != str:
-                raise TypeError('target must be of str type.')
-        if type(verbose) != int and type(verbose) != bool:
-            raise TypeError('verbose must be of int type or bool.')
         return nans.cleane(df, target, verbose)
 
 
 
 
-    # @staticmethod
     def encode_dataframe(self, 
-                        df : pd.DataFrame, 
-                        inplace : bool = False, 
-                        ) -> pd.DataFrame: 
+                         df : pd.DataFrame, 
+                         inplace : bool = False, 
+                         ) -> pd.DataFrame: 
         '''
         encode a dataframe
         Use case:
         df = Snaplib().encode_dataframe(df)
         '''
 
-        if not isinstance(df, pd.core.frame.DataFrame):
-            raise TypeError('The df must be a pandas.core.frame.DataFrame instance.')
-        if not isinstance(inplace, bool):
-            raise TypeError('The df must be a bool type instance.')
-
-
         if inplace is False:
             df_copied = df.copy()
 
@@ -135,22 +105,16 @@ class Snaplib__222:
 
         
 
-    # @staticmethod
     def decode_dataframe(self, 
-                        df : pd.DataFrame, 
-                        inplace : bool = False, 
-                        ) -> pd.DataFrame: 
+                         df : pd.DataFrame, 
+                         inplace : bool = False, 
+                         ) -> pd.DataFrame: 
         '''
         encode a dataframe
         Use case:
         df = Snaplib().decode_dataframe(df)
         '''
 
-        if not isinstance(df, pd.core.frame.DataFrame):
-            raise TypeError('The df must be a pandas.core.frame.DataFrame instance.')
-        if not isinstance(inplace, bool):
-            raise TypeError('The df must be a bool type instance.')
-
         if inplace is False:
             df_copied = df.copy()
             return self.ENCODER.decode_dataframe(df_copied)
@@ -165,7 +129,6 @@ class Snaplib__222:
                       target_name : str, 
                       k : int, 
                       ) -> dict: 
-
         '''
         Return a dictionary list of DataFrames and Series for target
         with next structure:
@@ -180,14 +143,6 @@ class Snaplib__222:
         Use case:
         k_fold_dict_data = Snaplib().k_folds_split( df, target_name, k)
         '''
-
-        if not isinstance(df, pd.core.frame.DataFrame):
-            raise TypeError('The df must be a pandas.core.frame.DataFrame instance.')
-        if type(target_name) != str:
-            raise TypeError('target_name must be of str type.')
-        if type(k) != int:
-            raise TypeError('k must be of int type.')
-
         return splitter.k_folds_split(df, target_name, k)
 
 
@@ -225,18 +180,6 @@ class Snaplib__222:
 
         TESTS on https://www.kaggle.com/artyomkolas/train-test-split-balanced-custom-in-snaplib/notebook
         '''
-
-        if not isinstance(df, pd.core.frame.DataFrame):
-            raise TypeError('The df must be a pandas.core.frame.DataFrame instance.')
-        if type(target_feature) != str:
-            raise TypeError('target_feature must be of str type.')
-        if type(test_size) != float:
-            raise TypeError('test_size must be of float type in [0.0, 1.0].')
-        if type(random_state) != int:
-            raise TypeError('random_state must be of int type.')
-        if type(research) != bool:
-            raise TypeError('research must be of bool type.')
-
         return  splitter.train_test_split_balanced( df, 
                                                     target_feature, 
                                                     test_size, 
@@ -246,6 +189,82 @@ class Snaplib__222:
 
 
 
+    @staticmethod
+    def cross_val(algorithms : list, 
+                  k_fold_dict : dict, 
+                  metric : Callable, 
+                  task : str, 
+                  cv : int, 
+                  verbose : int | bool = 0, 
+                  early_stopping_rounds : int = 0, 
+                  ) -> float:
+
+        ''' 
+        Cross Validation method for list of algorithms.
+        
+        Use case:
+        score = Snaplib().cross_val(algorithms, k_fold_dict, metric, task, cv, verbose=0):
+        
+        algorithms is a list of algorithms like algs = [
+                                                        [LGBMClassifier, dict(params)],
+                                                        [XGBClassifier, dict(params)], 
+                                                        [CatBoostClassifier, dict(params)],
+                                                        ]
+        k_fold_dict is a dictionary with the structure:
+        
+        K_FOLD = 3
+        k_fold_dict = { 
+                        'train_X' : [df_0, df_1, df_2],
+                        'test_X'  : [df_0, df_1, df_2],
+                        'train_y' : [seri_0, seri_1, seri_2],
+                        'test_y'  : [seri_0, seri_1, seri_2],
+                       }
+
+        Get k_fold_dict:
+        k_fold_dict = Snaplib().k_folds_split(df, target_name_str, K_FOLD)
+              
+        metric is a metric like f1_score or mean_absolute_error.
+        task='clsf' or 'regr', classification or regression.
+        cv is num K_FOLD integer 
+        verbose = 0 mute, 1 verbose.
+        early_stopping_rounds default 0 or positive int.
+
+        '''
+        return cross_validation.cross_val(algorithms, 
+                                          k_fold_dict, 
+                                          metric, 
+                                          task, 
+                                          cv, 
+                                          verbose=verbose, 
+                                          early_stopping_rounds=early_stopping_rounds)
+
+
+
+
+
+
+    def recover_data(df_in : pd.DataFrame, 
+                     device : str = 'cpu',
+                     verbose : int | bool = True,
+                     discrete_columns : list | str = 'auto', 
+                     ) -> pd.DataFrame:
+        ''' 
+        Imputing of missing values (np.nan) in tabular data, not TimeSeries.
+
+        Use case:
+        df = Snaplib().recover_data(df, verbose=True, stacking=True)
+        device must be "cpu" or "gpu". Sometime small datasets work faster with cpu.
+        if set verbose = if True algorithm runs cross validation tests and print results of tests for decision making.
+        discrete_columns = ['col_name_1', 'col_name_2', 'col_name_3', 'etc']
+
+        TESTS on https://www.kaggle.com/code/artyomkolas/nan-prediction-in-progress/notebook
+        '''
+        return recover.recover_data(df_in, 
+                                    device=device,
+                                    verbose=verbose,
+                                    discrete_columns=discrete_columns, 
+                                    )
+
 
 
 
@@ -671,7 +690,8 @@ class Snaplib:
                             X_test, 
                             y_test=None, 
                             task='clsf', 
-                            verbose=0
+                            verbose=0,
+                            early_stopping_rounds = 0, 
                             ):
         
         ''' 
@@ -679,14 +699,15 @@ class Snaplib:
         
         Use case:
         y_hat = Snaplib().fit_predict_stacked(
-                                        algorithms_list, 
-                                        X_train, 
-                                        y_train, 
-                                        X_test, 
-                                        y_test or None, 
-                                        task='clsf' or 'regr', 
-                                        verbose= 0, 1, 2
-                                        ):
+                                    algorithms_list, 
+                                    X_train, 
+                                    y_train, 
+                                    X_test, 
+                                    y_test or None, 
+                                    task='clsf' or 'regr', 
+                                    verbose= 0, 1,
+                                    early_stopping_rounds=0 or positve int
+                                    ):
         
         algorithms_list = list of algorithms [LGBMClassifier(), XGBClassifier(), CatBoostClassifier()].
         X_train and y_train are data for training list of algorithms.
@@ -716,7 +737,8 @@ class Snaplib:
 
         if type(verbose) != int and type(verbose) != bool:
             raise TypeError('verbose must be of int type or bool.')
-
+        if type(early_stopping_rounds) != int:
+            raise TypeError('early_stopping_rounds must be of int type.')
 
         stacked_predicts = pd.DataFrame()
         alg_names = []
@@ -728,9 +750,17 @@ class Snaplib:
             if alg_name in alg_names:
                 alg_name = alg_name + '_' + str(increment)
                 increment+=1
-            if y_test is not None and alg_name in ['LGB', 'XGB', 'Cat']:
+
+            if y_test is not None and \
+                alg_name in ['LGB', 'XGB', 'Cat'] and \
+                early_stopping_rounds:
+
                 model = alg(**params)
-                model.fit(X_train, y_train, eval_set=[(X_test, y_test)], early_stopping_rounds=200, verbose=False)
+                model.fit(X_train, 
+                          y_train, 
+                          eval_set=[(X_test, y_test)], 
+                          early_stopping_rounds=200, 
+                          verbose=False)
             else:
                 model = alg(**params)
                 model.fit(X_train, y_train)
@@ -762,9 +792,8 @@ class Snaplib:
         
     
 
-    def cross_val(self, algorithms, k_fold_dict, metric, task, cv, verbose=0):
+    def cross_val(self, algorithms, k_fold_dict, metric, task, cv, verbose=0, early_stopping_rounds=0):
 
-                
         ''' 
         Cross Validation method for list of algorithms.
         
@@ -793,6 +822,7 @@ class Snaplib:
         task='clsf' or 'regr', classification or regression.
         cv is num K_FOLD integer 
         verbose = 0 mute, 1 verbose.
+        early_stopping_rounds default 0 or positive int.
 
         '''
 
@@ -808,17 +838,19 @@ class Snaplib:
             raise TypeError('cv must be of int type.')
         if type(verbose) != int and type(verbose) != bool:
             raise TypeError('verbose must be of int type or bool.')
+        if type(early_stopping_rounds) != int:
+            raise TypeError('early_stopping_rounds must be of int type.')
 
 
         global num_combinations
         global all_prediction_df
-        global counter
+        global counter_cv
         global cpu_num
         
         
         results=[]    
         all_prediction_df = pd.DataFrame()
-        counter = self.Counter()
+        counter_cv = self.Counter()
         research_best_score=dict()
         cpu_num = cpu_count()
         num_combinations = 0
@@ -831,13 +863,14 @@ class Snaplib:
         
         
         for k in range(0, cv):
-            pred_frame = self.fit_predict_stacked(algorithms, 
+            pred_frame = self.fit_predict_stacked(  algorithms, 
                                                     k_fold_dict['train_X'][k], 
                                                     k_fold_dict['train_y'][k], 
                                                     k_fold_dict['test_X'][k], 
                                                     k_fold_dict['test_y'][k], 
                                                     task,
-                                                    verbose,
+                                                    verbose, 
+                                                    early_stopping_rounds=early_stopping_rounds,
                                                     )
 
             pred = pred_frame.loc[:, 'Y_HAT_STACKED']
@@ -998,14 +1031,14 @@ class Snaplib:
                     
         del num_combinations
         del all_prediction_df
-        del counter
+        del counter_cv
         del cpu_num
 
         return score
         
     @staticmethod
     def get_score_clsf(args):
-        print(f'\r{counter(cpu_num)} / {num_combinations} processed', end='', flush=True)
+        print(f'\r{counter_cv(cpu_num)} / {num_combinations} processed', end='', flush=True)
         algs_list = list(args[0])
         metric = args[1]
         y_hat_set = all_prediction_df.loc[:, algs_list].mode(axis=1)[0].astype('int64')
@@ -1014,7 +1047,7 @@ class Snaplib:
     
     @staticmethod
     def get_score_regr(args):
-        print(f'\r{counter(cpu_num)} / {num_combinations} processed', end='', flush=True)
+        print(f'\r{counter_cv(cpu_num)} / {num_combinations} processed', end='', flush=True)
         algs_list = list(args[0])
         metric = args[1]
         y_hat_set = all_prediction_df.loc[:, algs_list].mean(axis=1)
@@ -1034,9 +1067,10 @@ class Snaplib:
                     algorithms_list, 
                     X_train, 
                     y_train, 
-                    # X_val=None, 
-                    # y_val=None, 
-                    verbose=0
+                    X_val=None, 
+                    y_val=None, 
+                    verbose=0, 
+                    early_stopping_rounds=0, 
                     ):
         
         ''' 
@@ -1068,23 +1102,37 @@ class Snaplib:
         if not isinstance(y_train, pd.core.frame.Series) and not isinstance(y_train, np.ndarray):
             raise TypeError('The y__train must be a pandas.core.frame.Series instance or numpy.ndarray.')
 
-        # if X_val is not None:
-        #     if not isinstance(X_val, pd.core.frame.DataFrame):
-        #         raise TypeError('The X__val must be a pandas.core.frame.DataFrame instance.')
-        # if y_val is not None:
-        #     if not isinstance(y_val, pd.core.frame.Series) and not isinstance(y_val, np.ndarray):
-        #         raise TypeError('The y__val must be a pandas.core.frame.Series instance or numpy.ndarray.')
+        if X_val is not None:
+            if not isinstance(X_val, pd.core.frame.DataFrame):
+                raise TypeError('The X__val must be a pandas.core.frame.DataFrame instance.')
+        if y_val is not None:
+            if not isinstance(y_val, pd.core.frame.Series) and not isinstance(y_val, np.ndarray):
+                raise TypeError('The y__val must be a pandas.core.frame.Series instance or numpy.ndarray.')
 
         if type(verbose) != int and type(verbose) != bool:
             raise TypeError('verbose must be of int type or bool.')
+        if type(early_stopping_rounds) != int:
+            raise TypeError('early_stopping_rounds must be of int type.')
 
 
+        if y_val is not None:
+            X = pd.concat([X_train, X_val], ignore_index=True)
+            y = pd.concat([y_train, y_val], ignore_index=True)
+
         for alg in algorithms_list:
-            # alg_name = alg.__class__.__name__[:3]
-            # if X_val is not None and y_val is not None and alg_name in ['LGB', 'XGB', 'Cat']:
-            #     alg.fit(X_train, y_train, eval_set=[(X_val, y_val)], early_stopping_rounds=10, verbose=verbose)
-            # else:
-            alg.fit(X_train, y_train)
+            alg_name = alg.__class__.__name__[:3]
+            if X_val is not None and \
+                y_val is not None and \
+                alg_name in ['LGB', 'XGB', 'Cat'] and \
+                early_stopping_rounds:
+
+                alg.fit(X_train, 
+                        y_train, 
+                        eval_set=[(X_val, y_val)], 
+                        early_stopping_rounds=early_stopping_rounds, 
+                        verbose=verbose)
+            else:
+                alg.fit(X, y)
 
         return algorithms_list
     
